from pathlib import Path
from ursus.config import config
import logging
import re


logger = logging.getLogger(__name__)


class Linter():
    def lint(self, file_path: Path):
        """
        Lints the content for errors.
        """
        raise NotImplementedError


class LineLinter(Linter):
    """
    Lints a text file line by line.
    """
    def lint(self, file_path: Path):
        if self.file_suffixes and file_path.suffix.lower() not in self.file_suffixes:
            return

        with (config.content_path / file_path).open() as file:
            for line_no, line in enumerate(file.readlines()):
                for error, level in self.lint_line(file_path, line):
                    yield line_no, error, level

    def lint_line(self, file_path: Path, line: str):
        raise NotImplementedError


class RegexLinter(LineLinter):
    file_suffixes = None
    regex = re.compile(r'')

    def lint_line(self, file_path: Path, line: str):
        for match in self.regex.finditer(line):
            yield from self.handle_match(file_path, match)

    def handle_match(self, file_path: Path, match: re.Match):
        raise NotImplementedError



from ursus.utils import get_files_in_path
from ursus.config import config


class ContextProcessor:
    def process(self, context: dict, changed_files: set = None) -> dict:
        """Transforms the context and returns it. The context is used to render templates.

        Args:
            context (dict): An object that represents all data used to render templates
                (website info, blog posts, utility functions, etc.)
            changed_files (set, optional): A list of files that changed since the last context update.

        Returns:
            dict: The updated context
        """
        return context


class EntryContextProcessor(ContextProcessor):
    def process(self, context: dict, changed_files: set = None) -> dict:
        for file_path in get_files_in_path(config.content_path, changed_files):
            entry_uri = str(file_path)
            context['entries'].setdefault(entry_uri, {})
            self.process_entry(context, entry_uri)
        return context

    def process_entry(self, context: dict, entry_uri: str):
        raise NotImplementedError



from ursus.utils import import_class, get_files_in_path, log_colors
from ursus.config import config
from watchdog.observers import Observer
import logging
import time


def build(watch_for_changes: bool = False):
    """Runs ursus and builds a static website

    Args:
        watch_for_changes (bool, optional): Keep running, and rebuild when content or templates change
    """
    generator = import_class(config.generator)()

    if watch_for_changes:
        observer = Observer()

        try:
            generator.generate()
        except:
            logging.exception("Could not generate site")

        for path in generator.get_watched_paths():
            observer.schedule(generator.get_observer_event_handler(), path, recursive=True)
        observer.start()
        try:
            while True:
                time.sleep(1)
        finally:
            observer.stop()
            observer.join()
    else:
        generator.generate()


def lint(files_to_lint=None):
    """Lints the content for errors"""
    linters = [import_class(linter_path)() for linter_path in config.linters]

    for file_path in sorted(get_files_in_path(config.content_path, whitelist=files_to_lint)):
        logging.info(f"\033[1mLinting {str(file_path)}\033[0m")
        for linter in linters:
            for line_no, message, level in linter.lint(file_path):
                if line_no:
                    logging.log(level, f"{log_colors[level]}:{line_no}\033[0m {message}")
                else:
                    logging.log(level, f"{message}")



from itertools import chain
from markdown.blockprocessors import HashHeaderProcessor
from pathlib import Path
from requests.exceptions import ConnectionError
from urllib.parse import unquote, urlparse
from ursus.config import config
from ursus.context_processors.markdown import patched_slugify
from ursus.linters import RegexLinter
import logging
import re
import requests


class MarkdownLinksLinter(RegexLinter):
    """
    Verifies external links in Markdown files
    """
    file_suffixes = ('.md',)

    # Matches [], supports escaped brackets, ignores ![] images.
    first_half = r"(?P<first_half>!?\[(?P<text>([^\]]|\\\])*)(?<!\\)\])"

    # Matches (), supports escaped parentheses
    second_half = r"(?P<second_half>\((?P<url_group>([^\)]|\\\))*)(?<!\\)\))"

    regex = re.compile(first_half + second_half)

    def handle_match(self, file_path: Path, match: re.Match):
        text = match['text'].strip()
        url = None
        title = None
        is_image = match['first_half'].startswith('!')

        if match['url_group']:
            parts = match['url_group'].split(" ", maxsplit=1)
            url = parts[0]
            if len(parts) == 2:
                title = parts[1]

        for error, level in chain(
            self.validate_link_text(text, is_image, file_path),
            self.validate_link_title(title, is_image, file_path),
            self.validate_link_url(url, is_image, file_path),
        ):
            yield f"{error}: {match.group(0)}", level

    def validate_link_text(self, text: str, is_image: bool, file_path: Path):
        return
        yield

    def validate_link_url(self, url: str, is_image: bool, file_path: Path):
        return
        yield

    def validate_link_title(self, title: str, is_image: bool, file_path: Path):
        return
        yield


class MarkdownLinkTextsLinter(MarkdownLinksLinter):
    def validate_link_text(self, text: str, is_image: bool, file_path: Path):
        if not text:
            yield "Image has no alt text", logging.WARNING


class MarkdownLinkTitlesLinter(MarkdownLinksLinter):
    def validate_link_title(self, title: str, is_image: bool, file_path: Path):
        if title is not None:
            if not (title.startswith('"') and title.endswith('"')):
                yield "Title is not quoted", logging.ERROR
            elif title == title.strip('"'):
                yield "Title is empty", logging.WARNING


class MarkdownExternalLinksLinter(MarkdownLinksLinter):
    user_agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.response_cache = {}

    def validate_link_url(self, url: str, is_image: bool, file_path: Path):
        if not url:
            yield "Missing URL", logging.ERROR
        elif not url.startswith(('/', '#', 'http://', 'https://', 'mailto:', 'tel:')):
            yield "Relative or invalid URL", logging.WARNING
        else:
            if url.startswith(('http://', 'https://')):
                try:
                    cleaned_url = self.unescape_url(url)
                    if cleaned_url not in self.response_cache:
                        self.response_cache[cleaned_url] = requests.get(cleaned_url, timeout=5, headers={
                            'User-Agent': self.user_agent
                        })
                    response = self.response_cache[cleaned_url]
                    status_code = response.status_code
                except ConnectionError:
                    yield f"Connection error: {cleaned_url}", logging.ERROR
                except requests.exceptions.RequestException as exc:
                    yield f"URL {type(exc).__name__}: {cleaned_url}", logging.ERROR
                else:
                    if status_code in (404, 410):
                        yield f"URL returns HTTP {status_code}", logging.ERROR
                    elif status_code >= 400:
                        level = logging.WARNING if status_code in (403, 503) else logging.ERROR
                        yield f"URL returns HTTP {status_code}", level
                    elif response.history and response.history[-1].status_code == 301:
                        yield f"URL redirects to {response.url}", logging.INFO

    def escape_url(self, url: str) -> str:
        return url.replace('(', '\\(').replace(')', '\\)')

    def unescape_url(self, url: str) -> str:
        return url.replace('\\(', '(').replace('\\)', ')')


class MarkdownInternalLinksLinter(MarkdownLinksLinter):
    """
    Verify that internal links point to existing entries. If the URL has a fragment,
    it should point to an existing title fragment.
    """
    header_regex = HashHeaderProcessor.RE
    ignored_urls = (
        # re.compile(r'^/ignored-url$')
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.title_slugs_cache = {}

    def get_title_slugs(self, file_path: Path):
        if file_path not in self.title_slugs_cache:
            self.title_slugs_cache[file_path] = set()
            with file_path.open() as file:
                for line in file.readlines():
                    if bool(self.header_regex.search(line)):
                        self.title_slugs_cache[file_path].add(
                            patched_slugify(line.lstrip('#').strip(), '-')
                        )
        return self.title_slugs_cache[file_path]

    def validate_link_url(self, url: str, is_image: bool, current_file_path: Path):
        for ignored_url in self.ignored_urls:
            if ignored_url.match(url):
                return

        url_parts = urlparse(url)
        title_slug = url_parts.fragment

        # Convert relative and absolute URLs to content paths
        if config.site_url and url.startswith(config.site_url):
            file_path = config.content_path / unquote(url_parts.path.removeprefix(config.site_url))
        elif url.startswith('/'):
            file_path = config.content_path / unquote(url_parts.path.lstrip('/'))
        elif url.startswith('#'):
            file_path = config.content_path / current_file_path
        elif not url_parts.scheme and not url_parts.netloc:  # Relative URL
            file_path = current_file_path.parent / unquote(url_parts.path)
        else:
            return

        if file_path.suffix.lower() == config.html_url_extension:
            file_path = file_path.with_suffix('.md')

        if not file_path.exists():
            yield "Entry not found", logging.ERROR
        elif title_slug and title_slug not in self.get_title_slugs(file_path):
            yield "URL fragment not found", logging.ERROR



from datetime import datetime
from pathlib import Path
from ursus.config import config
from ursus.context_processors import ContextProcessor
import git


class GitDateProcessor(ContextProcessor):
    """
    Sets entry.date_updated to the date of the latest commit.
    """
    def __init__(self):
        super().__init__()
        if not config.fast_rebuilds:
            self.repo = git.Repo(config.content_path, search_parent_directories=True)
            self.repo_root = Path(self.repo.working_dir)

    def commit_path_to_entry_uri(self, commit_path: str):
        abs_commit_path = self.repo_root / commit_path
        try:
            return str(abs_commit_path.relative_to(config.content_path))
        except ValueError:
            return None

    def process(self, context: dict, changed_files: set = None) -> dict:
        if not config.fast_rebuilds:
            for commit in self.repo.iter_commits("master"):
                commit_date = datetime.fromtimestamp(commit.authored_date)
                for file in self.repo.git.show(commit.hexsha, name_only=True, diff_filter='ACMR').split('\n'):
                    entry_uri = self.commit_path_to_entry_uri(file)
                    if entry_uri and entry_uri in context['entries']:
                        entry = context['entries'][entry_uri]
                        if not entry.get('date_updated') or commit_date > entry['date_updated']:
                            entry['date_updated'] = commit_date

        for entry in context['entries'].values():
            entry.setdefault('date_updated', datetime.now())
        return context



from setuptools import setup, find_packages
from pathlib import Path
long_description = (Path(__file__).parent / "README.md").read_text()

setup(
    name='ursus_ssg',
    version='1.1.0',
    description='Static site generator',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='http://github.com/nicbou/ursus',
    author='Nicolas Bouliane',
    author_email='contact@nicolasbouliane.com',
    license='MIT',
    packages=find_packages(),
    scripts=['bin/ursus'],
    python_requires='>=3.11',
    install_requires=[
        'coloredlogs==15.0.1',
        'GitPython==3.1.30',
        'imagesize==1.4.1',
        'Jinja2==3.1.2',
        'jinja2-simple-tags==0.5.0',
        'libsass==0.22.0',
        'lunr==0.6.2',
        # 'Markdown==3.4.3',
        # Use main branch when github.com/Python-Markdown/markdown/pull/1354/ is merged
        'Markdown @ git+https://github.com/nicbou/markdown.git@nicbou-patch-1',
        'MarkupSafe==2.1.1',
        'ordered-set==4.1.0',
        'PyMuPDF==1.21.1',
        'Pillow==9.4.0',  # Pillow-simd is much faster, but requires binaries to be installed separately.
        'watchdog==2.2.1',
        'requests==2.31.0',
        'rjsmin==1.2.1',
        'rcssmin==1.1.1',
    ],
    zip_safe=False,
)
